"""Configuration for imitation.scripts.train_preference_comparisons."""

import sacred
from torch import nn

from imitation.algorithms import preference_comparisons
from imitation.scripts.ingredients import environment
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients import policy_evaluation, reward, rl

# Note: All the hyperparameter configs in the file are of the tuned
# hyperparameters of the RL algorithm of the respective environment.
# Taken from imitation/scripts/config/train_rl.py

train_preference_comparisons_ex = sacred.Experiment(
    "train_preference_comparisons",
    ingredients=[
        logging_ingredient.logging_ingredient,
        environment.environment_ingredient,
        reward.reward_ingredient,
        rl.rl_ingredient,
        policy_evaluation.policy_evaluation_ingredient,
    ],
)


MUJOCO_SHARED_LOCALS = dict(rl=dict(rl_kwargs=dict(ent_coef=0.1)))
ANT_SHARED_LOCALS = dict(
    total_timesteps=int(3e7),
    rl=dict(batch_size=16384),
)


@train_preference_comparisons_ex.config
def train_defaults():
    fragment_length = 100  # timesteps per fragment used for comparisons
    total_timesteps = int(1e6)  # total number of environment timesteps
    total_comparisons = 5000  # total number of comparisons to elicit
    num_iterations = 5  # Arbitrary, should be tuned for the task
    comparison_queue_size = None
    # factor by which to oversample transitions before creating fragments
    transition_oversampling = 1
    # fraction of total_comparisons that will be sampled right at the beginning
    initial_comparison_frac = 0.1
    # fraction of sampled trajectories that will include some random actions
    exploration_frac = 0.0
    preference_model_kwargs = {}
    reward_trainer_kwargs = {
        "epochs": 3,
    }
    save_preferences = False  # save preference dataset at the end?
    agent_path = None  # path to a (partially) trained agent to load at the beginning
    # type of PreferenceGatherer to use
    gatherer_cls = preference_comparisons.SyntheticGatherer
    # arguments passed on to the PreferenceGatherer specified by gatherer_cls
    gatherer_kwargs = {}
    active_selection = False
    active_selection_oversampling = 2
    uncertainty_on = "logit"
    fragmenter_kwargs = {
        "warning_threshold": 0,
    }
    # path to a pickled sequence of trajectories used instead of training an agent
    trajectory_path = None
    trajectory_generator_kwargs = {}  # kwargs to pass to trajectory generator
    allow_variable_horizon = False

    checkpoint_interval = 0  # Num epochs between saving (<0 disables, =0 final only)
    query_schedule = "hyperbolic"


@train_preference_comparisons_ex.named_config
def cartpole():
    environment = dict(gym_id="CartPole-v1")
    allow_variable_horizon = True


@train_preference_comparisons_ex.named_config
def seals_ant():
    environment = dict(gym_id="seals/Ant-v0")
    rl = dict(
        batch_size=2048,
        rl_kwargs=dict(
            batch_size=16,
            clip_range=0.3,
            ent_coef=3.1441389214159857e-06,
            gae_lambda=0.8,
            gamma=0.995,
            learning_rate=0.00017959211641976886,
            max_grad_norm=0.9,
            n_epochs=10,
            # policy_kwargs are same as the defaults
            vf_coef=0.4351450387648799,
        ),
    )


@train_preference_comparisons_ex.named_config
def half_cheetah():
    locals().update(**MUJOCO_SHARED_LOCALS)
    environment = dict(gym_id="HalfCheetah-v2")
    rl = dict(batch_size=16384, rl_kwargs=dict(batch_size=1024))


@train_preference_comparisons_ex.named_config
def seals_half_cheetah():
    environment = dict(gym_id="seals/HalfCheetah-v0")
    rl = dict(
        batch_size=512,
        rl_kwargs=dict(
            batch_size=64,
            clip_range=0.1,
            ent_coef=3.794797423594763e-06,
            gae_lambda=0.95,
            gamma=0.95,
            learning_rate=0.0003286871805949382,
            max_grad_norm=0.8,
            n_epochs=5,
            vf_coef=0.11483689492120866,
        ),
    )
    num_iterations = 50
    total_timesteps = 20000000


@train_preference_comparisons_ex.named_config
def seals_hopper():
    environment = dict(gym_id="seals/Hopper-v0")
    policy = dict(
        policy_cls="MlpPolicy",
        policy_kwargs=dict(
            activation_fn=nn.ReLU,
            net_arch=[dict(pi=[64, 64], vf=[64, 64])],
        ),
    )
    rl = dict(
        batch_size=2048,
        rl_kwargs=dict(
            batch_size=512,
            clip_range=0.1,
            ent_coef=0.0010159833764878474,
            gae_lambda=0.98,
            gamma=0.995,
            learning_rate=0.0003904770450788824,
            max_grad_norm=0.9,
            n_epochs=20,
            vf_coef=0.20315938606555833,
        ),
    )


@train_preference_comparisons_ex.named_config
def seals_swimmer():
    environment = dict(gym_id="seals/Swimmer-v0")
    policy = dict(
        policy_cls="MlpPolicy",
        policy_kwargs=dict(
            activation_fn=nn.ReLU,
            net_arch=[dict(pi=[64, 64], vf=[64, 64])],
        ),
    )
    rl = dict(
        batch_size=2048,
        rl_kwargs=dict(
            batch_size=64,
            clip_range=0.1,
            ent_coef=5.167107294612664e-08,
            gae_lambda=0.95,
            gamma=0.999,
            learning_rate=0.000414936134792374,
            max_grad_norm=2,
            n_epochs=5,
            # policy_kwargs are same as the defaults
            vf_coef=0.6162112311062333,
        ),
    )


@train_preference_comparisons_ex.named_config
def seals_walker():
    environment = dict(gym_id="seals/Walker2d-v0")
    policy = dict(
        policy_cls="MlpPolicy",
        policy_kwargs=dict(
            activation_fn=nn.ReLU,
            net_arch=[dict(pi=[64, 64], vf=[64, 64])],
        ),
    )
    rl = dict(
        batch_size=8192,
        rl_kwargs=dict(
            batch_size=128,
            clip_range=0.4,
            ent_coef=0.00013057334805552262,
            gae_lambda=0.92,
            gamma=0.98,
            learning_rate=0.000138575372312869,
            max_grad_norm=0.6,
            n_epochs=20,
            # policy_kwargs are same as the defaults
            vf_coef=0.6167177795726859,
        ),
    )


@train_preference_comparisons_ex.named_config
def seals_humanoid():
    locals().update(**MUJOCO_SHARED_LOCALS)
    environment = dict(gym_id="seals/Humanoid-v0")
    total_timesteps = int(4e6)


@train_preference_comparisons_ex.named_config
def seals_cartpole():
    environment = dict(gym_id="seals/CartPole-v0")


@train_preference_comparisons_ex.named_config
def pendulum():
    environment = dict(gym_id="Pendulum-v1")


@train_preference_comparisons_ex.named_config
def mountain_car():
    environment = dict(gym_id="MountainCar-v0")
    allow_variable_horizon = True


@train_preference_comparisons_ex.named_config
def seals_mountain_car():
    environment = dict(gym_id="seals/MountainCar-v0")


@train_preference_comparisons_ex.named_config
def fast():
    # Minimize the amount of computation. Useful for test cases.
    total_timesteps = 50
    total_comparisons = 5
    initial_comparison_frac = 0.2
    num_iterations = 1
    fragment_length = 2
    reward_trainer_kwargs = {
        "epochs": 1,
    }



"""Utilities for processing the config files of the benchmarking runs."""

import argparse
import pathlib
from typing import List


def filter_config_files(files: List[str], /) -> List[pathlib.Path]:
    """Filter config files.

    Filter the list of config files to ignore the seed, and only return the last
    file per algorithm and environment.

    Args:
        files: list of config files

    Returns:
        list of config files
    """
    config_files = [pathlib.Path(config_file) for config_file in files]
    config_files = [
        config_file for config_file in config_files if config_file.name == "config.json"
    ]
    # all files have the following format:
    # [/base/path/to/file]/<algo_and_environment>/<other_info>/sacred/1/config.json
    experiments = {}
    for config_file in config_files:
        experiment = config_file.parents[3]
        if experiment not in experiments:
            experiments[experiment] = [config_file]
        else:
            experiments[experiment].append(config_file)
    final_config_files = []
    for experiment, config_files in experiments.items():
        config_files.sort(key=lambda config_file: config_file.parents[1])
        final_config_files.append(config_files[-1])
    return final_config_files


def remove_empty_dicts(d: dict):
    """Remove empty dictionaries in place.

    This is a recursive function that will remove empty dictionaries from a dictionary.

    Args:
        d: dictionary to be filtered
    """
    for key, value in list(d.items()):
        if isinstance(value, dict):
            remove_empty_dicts(value)
            if not value:
                d.pop(key)
        elif value == {}:
            d.pop(key)


def clean_config_file(file: pathlib.Path, write_path: pathlib.Path, /) -> None:
    """Clean a config file.

    reads the file, loads from json to dict, removes keys related to e.g. seeds,
    config paths, leaving only hyperparameters, and writes back to file.

    Args:
        file: path to config file
        write_path: path to write the cleaned config file
    """
    import json

    with open(file) as f:
        config = json.load(f)
    # remove key 'agent_path'
    config.pop("agent_path")
    config.pop("seed")
    config.get("demonstrations", {}).pop("path")
    config.get("expert", {}).get("loader_kwargs", {}).pop("path", None)
    env_name = config.pop("environment").pop("gym_id")
    config["environment"] = {"gym_id": env_name}
    config.pop("show_config", None)

    remove_empty_dicts(config)
    # files are of the format
    # /path/to/file/<algo>_<env>_best_hp_eval/<other_info>/sacred/1/config.json
    # we want to write to /<write_path>/<algo>_<env>.json
    with open(write_path / f"{file.parents[3].name}.json", "w") as f:
        json.dump(config, f, indent=4)


def main():
    """Main function of the script."""
    # get two arguments from the terminal. The first positional argument contains the
    # path to a txt file that has a list of paths to config files.
    # The second positional argument is the path to the directory where
    # the cleaned config files should be written.

    parser = argparse.ArgumentParser()
    parser.add_argument("config_files", type=str)
    parser.add_argument("write_path", type=str)
    args = parser.parse_args()

    # read the list of config files from the txt file
    with open(args.config_files) as f:
        config_files_str = f.read().splitlines()
    config_files = [pathlib.Path(file) for file in config_files_str]

    write_path = pathlib.Path(args.write_path)

    # make sure the write path and all the config files already exist
    if not write_path.exists():
        raise ValueError(f"write path {write_path} does not exist")
    for file in config_files:
        if not file.exists():
            raise ValueError(f"config file {file} does not exist")

    for file in config_files:
        clean_config_file(file, write_path)


if __name__ == "__main__":
    main()



"""Converts old-style pickle or npz trajectories to new-style HuggingFace datasets.

See https://github.com/HumanCompatibleAI/imitation/pull/448 for a description
of the new trajectory format.

This script takes as command-line input multiple paths to saved trajectories,
in the old .pkl or .npz format. It then saves each sequence in the new HuggingFace
datasets format. The path is the same as the original but a directory without an
extension (i.e. "A.pkl" -> "A/", "A.npz" -> "A/", "A/" -> "A/", "A.foo" -> "A/").
"""

import pathlib
import warnings

from imitation.data import huggingface_utils, serialize, types
from imitation.util import util


def update_traj_file_in_place(path_str: types.AnyPath, /) -> pathlib.Path:
    """Converts pickle or npz file to the new HuggingFace format.

    The new data is saved as `Sequence[imitation.types.TrajectoryWithRew]`.

    Args:
        path_str: Path to a pickle or npz file containing
            `Sequence[imitation.types.Trajectory]`
            or `Sequence[imitation.old_types.TrajectoryWithRew]`.

    Returns:
        The path to the converted trajectory dataset.
    """
    path = util.parse_path(path_str)
    with warnings.catch_warnings():
        # Filter out DeprecationWarning because we expect to load old trajectories here.
        warnings.filterwarnings(
            "ignore",
            message="Loading old .* version of Trajectories.*",
            category=DeprecationWarning,
        )
        trajs = serialize.load(path)

    if isinstance(
        trajs,
        huggingface_utils.TrajectoryDatasetSequence,
    ):
        warnings.warn(f"File {path} is already in the new format. Skipping.")
        return path
    else:
        converted_path = path.with_suffix("")
        serialize.save(converted_path, trajs)
        return converted_path


def main():  # pragma: no cover
    import sys

    if len(sys.argv) <= 1:
        print("Supply at least one path to pickled trajectories.")
    else:
        for path in sys.argv[1:]:
            update_traj_file_in_place(path)


if __name__ == "__main__":
    main()



"""Helpers to convert between Trajectories and HuggingFace's datasets library."""
from typing import Any, Dict, Iterable, Optional, Sequence, cast

import datasets
import jsonpickle
import numpy as np

from imitation.data import types


class TrajectoryDatasetSequence(Sequence[types.Trajectory]):
    """A wrapper to present an HF dataset as a sequence of trajectories.

    Converts the dataset to a sequence of trajectories on the fly.
    """

    def __init__(self, dataset: datasets.Dataset):
        """Construct a TrajectoryDatasetSequence."""

        def numpy_transform(batch):
            # No need to convert infos to a numpy array.
            # This speeds up the conversion quite a lot
            return {
                key: np.asarray(val) if key != "infos" else val
                for key, val in batch.items()
            }

        # TODO: this is just a temporary workaround for
        #  https://github.com/huggingface/datasets/issues/5517
        #  switch to .with_format("numpy") once it's fixed
        self._dataset = dataset.with_transform(numpy_transform)
        self._trajectory_class = (
            types.TrajectoryWithRew if "rews" in dataset.features else types.Trajectory
        )

    def __len__(self) -> int:
        return len(self._dataset)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            # Note: we could use self._dataset[idx] here and then convert the result of
            #   that to a series of trajectories, but if we do that, we run into trouble
            #   with the custom numpy transform that we apply in the constructor.
            #   The transform is applied to the whole slice, which might contain
            #   trajectories of different lengths which is not supported by numpy.
            return [self[i] for i in range(*idx.indices(len(self)))]
        else:
            # Extract the trajectory kwargs from the dataset
            kwargs = self._dataset[idx]

            # Ensure that the infos are decoded lazily using jsonpickle
            kwargs["infos"] = _LazyDecodedList(kwargs["infos"])

            return self._trajectory_class(**kwargs)

    @property
    def dataset(self):
        """Return the underlying HF dataset."""
        # Note: since we apply the custom numpy transform in the constructor, we remove
        #   it again before returning the dataset. This ensures that the dataset is
        #   returned in the original format and can be saved to disk
        #   (the custom transform can not be saved to disk since it is not pickleable).
        return self._dataset.with_transform(None)


class _LazyDecodedList(Sequence[Any]):
    """A wrapper to lazily decode a list of jsonpickled strings.

    Decoded results are cached to avoid decoding the same string multiple times.

    This is used to decode the infos of a trajectory only when they are accessed.
    """

    def __init__(self, encoded_list: Sequence[str]):
        self._encoded_list = encoded_list
        self._decoded_cache: Dict[int, Any] = {}

    def __len__(self):
        return len(self._encoded_list)

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return [self[i] for i in range(*idx.indices(len(self)))]
        else:
            if idx not in self._decoded_cache:
                self._decoded_cache[idx] = jsonpickle.decode(self._encoded_list[idx])
            return self._decoded_cache[idx]


def trajectories_to_dict(
    trajectories: Sequence[types.Trajectory],
) -> Dict[str, Sequence[Any]]:
    """Convert a sequence of trajectories to a dict.

    The dict has the following fields:

    * obs: The observations. Shape: (num_trajectories, num_timesteps, obs_dim).
    * acts: The actions. Shape: (num_trajectories, num_timesteps, act_dim).
    * infos: The infos. Shape: (num_trajectories, num_timesteps) as jsonpickled str.
    * terminal: The terminal flags. Shape: (num_trajectories, num_timesteps, ).
    * rews: The rewards. Shape: (num_trajectories, num_timesteps) if applicable.

    This dict can be used to construct a HuggingFace dataset.

    Args:
        trajectories: The trajectories to save.

    Raises:
        ValueError: If not all trajectories have the same type, i.e. some are
            `Trajectory` and others are `TrajectoryWithRew`.

    Returns:
        A dict representing the trajectories.
    """
    # Check that all trajectories have rewards or none have rewards
    has_reward = [isinstance(traj, types.TrajectoryWithRew) for traj in trajectories]
    all_trajectories_have_reward = all(has_reward)
    if not all_trajectories_have_reward and any(has_reward):
        raise ValueError("Some trajectories have rewards but not all")

    # Convert to dict
    trajectory_dict: Dict[str, Sequence[Any]] = dict(
        obs=[traj.obs for traj in trajectories],
        acts=[traj.acts for traj in trajectories],
        # Replace 'None' values for `infos`` with array of empty dicts
        infos=[
            traj.infos if traj.infos is not None else [{}] * len(traj)
            for traj in trajectories
        ],
        terminal=[traj.terminal for traj in trajectories],
    )
    if any(isinstance(traj.obs, types.DictObs) for traj in trajectories):
        raise ValueError("DictObs are not currently supported")

    # Encode infos as jsonpickled strings
    trajectory_dict["infos"] = [
        [jsonpickle.encode(info) for info in traj_infos]
        for traj_infos in cast(Iterable[Iterable[Dict]], trajectory_dict["infos"])
    ]

    # Add rewards if applicable
    if all_trajectories_have_reward:
        trajectory_dict["rews"] = [
            cast(types.TrajectoryWithRew, traj).rews for traj in trajectories
        ]
    return trajectory_dict


def trajectories_to_dataset(
    trajectories: Sequence[types.Trajectory],
    info: Optional[datasets.DatasetInfo] = None,
) -> datasets.Dataset:
    """Convert a sequence of trajectories to a HuggingFace dataset."""
    if isinstance(trajectories, TrajectoryDatasetSequence):
        return trajectories.dataset
    else:
        return datasets.Dataset.from_dict(trajectories_to_dict(trajectories), info=info)
